/*
 * Copyright (C) 2006-2008 Andre Noll <maan@systemlinux.org>
 *
 * Licensed under the GPL v2. For licencing details see COPYING.
 */

/** \file osx_write.c paraslash's output plugin for MacOs */

/*
 * based on mosx-mpg123, by Guillaume Outters and Steven A. Kortze
 * <skortze@sourceforge.net>
 */

#include <sys/types.h>
#include <dirent.h>
#include <CoreAudio/CoreAudio.h>
#include "para.h"
#include "fd.h"
#include "string.h"
#include "list.h"
#include "sched.h"
#include "write.h"
#include "osx_write.cmdline.h"
#include "error.h"


#include <CoreAudio/CoreAudio.h>
#include <AudioUnit/AudioUnit.h>
#include <AudioToolbox/DefaultAudioOutput.h>

/** describes one input buffer for the osx writer */
struct osx_buffer {
	/** pointer to the beginning of the buffer */
	float *buffer;
	/** the size of this buffer */
	long size;
	/** current position in the buffer */
	float *ptr;
	/** number of floats not yet consuned */
	long remaining;
	/** pointer to the next audio buffer */
	struct osx_buffer *next;
};

/** data specific to the osx writer */
struct private_osx_write_data {
	/** the main control structure for audio data manipulation */
	AudioUnit audio_unit;
	/** non-zero if playback has started */
	char play;
	/** callback reads audio data from this buffer */
	struct osx_buffer *from;
	/** the post_select writes audio data here */
	struct osx_buffer *to;
	/** sample rate of the current audio stream */
	unsigned samplerate;
	/** number of channels of the current audio stream */
	unsigned channels;
};

static void destroy_buffers(struct private_osx_write_data *powd)
{
	struct osx_buffer *ptr;
	struct osx_buffer *ptr2;
	ptr = powd->to->next;
	powd->to->next = NULL;
	while (ptr) {
		ptr2 = ptr->next;
		free(ptr->buffer);
		free(ptr);
		ptr = ptr2;
	}
}

static void init_buffers(struct writer_node *wn)
{
	struct private_osx_write_data *powd = wn->private_data;
	struct osx_write_args_info *conf = wn->conf;
	struct osx_buffer **ptrptr;
	int i;

	ptrptr = &powd->to;
	for (i = 0; i < conf->numbuffers_arg; i++) {
		*ptrptr = malloc(sizeof(struct osx_buffer));
		(*ptrptr)->size = 0;
		(*ptrptr)->remaining = 0;
		(*ptrptr)->buffer = NULL;
		ptrptr = &(*ptrptr)->next;
	}
	*ptrptr = powd->from = powd->to;
}

static void fill_buffer(struct osx_buffer *b, short *source, long size)
{
	float *dest;

	if (b->remaining) /* Non empty buffer, must still be playing */
		return;
	if (b->size != size) {
		b->buffer = para_realloc(b->buffer, size * sizeof(float));
		b->size = size;
	}
	dest = b->buffer;
	while (size--)
		*dest++ = (*source++) / 32768.0;
	b->ptr = b->buffer;
	b->remaining = b->size;
}

static OSStatus osx_callback(void * inClientData,
	__a_unused AudioUnitRenderActionFlags *inActionFlags,
	__a_unused const AudioTimeStamp *inTimeStamp,
	__a_unused  UInt32 inBusNumber,
	__a_unused UInt32 inNumFrames,
	AudioBufferList *outOutputData)
{
	long m, n;
	float *dest;
	int i;
	struct private_osx_write_data *powd = inClientData;

//	PARA_INFO_LOG("%p\n", powd);
	for (i = 0; i < outOutputData->mNumberBuffers; ++i) {
		/* what we have to fill */
		m = outOutputData->mBuffers[i].mDataByteSize / sizeof(float);
		dest = outOutputData->mBuffers[i].mData;
		while (m > 0) {
			if ((n = powd->from->remaining) <= 0) {
				PARA_INFO_LOG("buffer underrun\n");
				return 0;
			}
//			PARA_INFO_LOG("buf %p: n = %ld, m= %ld\n", powd->from->buffer, n, m);
			/*
			 * we dump what we can. In fact, just the necessary
			 * should be sufficient
			 */
			if (n > m)
				n = m;
			memcpy(dest, powd->from->ptr, n * sizeof(float));
			dest += n;
			/* remember all done work */
			m -= n;
			powd->from->ptr += n;
			if ((powd->from->remaining -= n) <= 0)
				powd->from = powd->from->next;
		}
	}
	return 0;
}

static int osx_write_open(struct writer_node *wn)
{
	struct private_osx_write_data *powd = para_calloc(
		sizeof(struct private_osx_write_data));
	ComponentDescription desc;
	Component comp;
	AURenderCallbackStruct inputCallback = {osx_callback, powd};
	AudioStreamBasicDescription format;
	int ret;
	struct writer_node_group *wng = wn->wng;
	struct osx_write_args_info *conf = wn->conf;

	wn->private_data = powd;
	/* where did that default audio output go? */
	desc.componentType = kAudioUnitType_Output;
	desc.componentSubType = kAudioUnitSubType_DefaultOutput;
	/* NOTE: and if default output isn't Apple? */
	desc.componentManufacturer = kAudioUnitManufacturer_Apple;
	desc.componentFlags = 0;
	desc.componentFlagsMask = 0;
	ret = -E_DEFAULT_COMP;
	comp = FindNextComponent(NULL, &desc);
	if (!comp)
		goto e0;
	ret = -E_OPEN_COMP;
	if (OpenAComponent(comp, &powd->audio_unit))
		goto e0;
	ret = -E_UNIT_INIT;
	if (AudioUnitInitialize(powd->audio_unit))
		goto e1;
	powd->play = 0;
	/* Hmmm, let's choose PCM format */
	/* We tell the Output Unit what format we're going to supply data to it.
	 * This is necessary if you're providing data through an input callback
	 * AND you want the DefaultOutputUnit to do any format conversions
	 * necessary from your format to the device's format.
	 */
	if (!conf->samplerate_given && wng->samplerate)
		powd->samplerate = *wng->samplerate;
	else
		powd->samplerate = conf->samplerate_arg;
	format.mSampleRate = powd->samplerate;
	/* The specific encoding type of audio stream*/
	format.mFormatID = kAudioFormatLinearPCM;
	/* flags specific to each format */
	format.mFormatFlags = kLinearPCMFormatFlagIsFloat
		| kLinearPCMFormatFlagIsPacked
		| kLinearPCMFormatFlagIsBigEndian;
	if (!conf->channels_given && wng->channels)
		powd->channels = *wng->channels;
	else
		powd->channels = conf->channels_arg;
	format.mChannelsPerFrame = powd->channels;
	format.mFramesPerPacket = 1;
	format.mBytesPerPacket = format.mChannelsPerFrame * sizeof(float);
	format.mBytesPerFrame = format.mFramesPerPacket * format.mBytesPerPacket;
	/* one of the most constant constants of the whole computer history */
	format.mBitsPerChannel = sizeof(float) * 8;
	ret = -E_STREAM_FORMAT;
	if (AudioUnitSetProperty(powd->audio_unit, kAudioUnitProperty_StreamFormat,
			kAudioUnitScope_Input, 0, &format,
			sizeof(AudioStreamBasicDescription)))
		goto e2;
	init_buffers(wn);
	ret = -E_ADD_CALLBACK;
	if (AudioUnitSetProperty(powd->audio_unit, kAudioUnitProperty_SetRenderCallback,
			kAudioUnitScope_Input, 0, &inputCallback,
			sizeof(inputCallback)) < 0)
		goto e3;
	return 1;
e3:
	destroy_buffers(powd);
e2:
	AudioUnitUninitialize(powd->audio_unit);
e1:
	CloseComponent(powd->audio_unit);
e0:
	return ret;
}

__malloc static void *osx_write_parse_config(const char *options)
{
	struct osx_write_args_info *conf
		= para_calloc(sizeof(struct osx_write_args_info));
	PARA_INFO_LOG("options: %s\n", options);
	int ret = osx_cmdline_parser_string(options, conf, "osx_write");
	if (ret)
		goto err_out;
	return conf;
err_out:
	free(conf);
	return NULL;

}

static void osx_write_close(struct writer_node *wn)
{
	struct private_osx_write_data *powd = wn->private_data;

	PARA_INFO_LOG("closing writer node %p\n", wn);
	AudioOutputUnitStop(powd->audio_unit);
	AudioUnitUninitialize(powd->audio_unit);
	CloseComponent(powd->audio_unit);
	destroy_buffers(powd);
	free(powd);
}

static int need_new_buffer(struct writer_node *wn)
{
	struct writer_node_group *wng = wn->wng;
	struct private_osx_write_data *powd = wn->private_data;

	if (*wng->loaded < sizeof(short))
		return 0;
	if (powd->to->remaining) /* Non empty buffer, must still be playing */
		return 0;
	return 1;
}

static int osx_write_post_select(__a_unused struct sched *s,
		struct writer_node *wn)
{
	struct private_osx_write_data *powd = wn->private_data;
	struct writer_node_group *wng = wn->wng;
	short *data = (short*)wng->buf;

	if (!need_new_buffer(wn))
		return 1;
	fill_buffer(powd->to, data, *wng->loaded / sizeof(short));
	powd->to = powd->to->next;
	wn->written = *wng->loaded;
	if (!powd->play) {
		if (AudioOutputUnitStart(powd->audio_unit))
			return -E_UNIT_START;
		powd->play = 1;
	}
	return 1;
}

static int osx_write_pre_select(struct sched *s, __a_unused struct writer_node *wn)
{
	struct private_osx_write_data *powd = wn->private_data;
	struct writer_node_group *wng = wn->wng;
	size_t numbytes = powd->to->remaining * sizeof(short);
	struct timeval tmp = {.tv_sec = 1, .tv_usec = 0}, delay = tmp;
	unsigned long divisor;

	if (!numbytes && *wng->loaded >= sizeof(short))
		goto min_delay; /* there's a buffer to fill */
	if (!numbytes)
		return 1;
	divisor = powd->samplerate * powd->channels * 2 / numbytes;
	if (divisor)
		tv_divide(divisor, &tmp, &delay);
	if (tv_diff(&s->timeout, &delay, NULL) > 0)
		s->timeout = delay;
//	PARA_DEBUG_LOG("delay: %lu:%lu\n", (long unsigned) s->timeout.tv_sec,
//		(long unsigned) s->timeout.tv_usec);
	return 1;
min_delay:
	PARA_DEBUG_LOG("%s\n", "minimal delay");
	s->timeout.tv_sec = 0;
	s->timeout.tv_usec = 1;
	return 1;
}

/** the init function of the osx writer */
void osx_write_init(struct writer *w)
{
	w->open = osx_write_open;
	w->close = osx_write_close;
	w->pre_select = osx_write_pre_select;
	w->post_select = osx_write_post_select;
	w->parse_config = osx_write_parse_config;
	w->shutdown = NULL; /* nothing to do */
}
