/*
 * Copyright (C) 2006-2008 Andre Noll <maan@systemlinux.org>
 *
 * Licensed under the GPL v2. For licencing details see COPYING.
 */

/** \file write_common.c common functions of para_audiod and para_write */

#include "para.h"
#include "string.h"
#include "list.h"
#include "sched.h"
#include "write.h"
#include "error.h"

/** the array containing the names of all supported writers */
const char *writer_names[] ={WRITER_NAMES};

/** the array of supported writers */
struct writer writers[NUM_SUPPORTED_WRITERS] = {WRITER_ARRAY};

static void wng_pre_select(__a_unused struct sched *s, struct task *t)
{
	struct writer_node_group *g = t->private_data;
	int i;

	FOR_EACH_WRITER_NODE(i, g) {
		struct writer_node *wn = &g->writer_nodes[i];
		t->ret = wn->writer->pre_select(s, wn);
		if (t->ret < 0) {
			g->error = t->ret;
			return;
		}
	}
}

static void wng_post_select(struct sched *s, struct task *t)
{
	struct writer_node_group *g = t->private_data;
	int i;
	size_t min_written = 0;

	FOR_EACH_WRITER_NODE(i, g) {
		struct writer_node *wn = &g->writer_nodes[i];
		t->ret = wn->writer->post_select(s, wn);
		if (t->ret < 0) {
			g->error = t->ret;
			return;
		}
		if (!i)
			min_written = wn->written;
		else
			min_written = PARA_MIN(min_written, wn->written);
	}
//	PARA_INFO_LOG("loaded: %zd, min_written: %zd bytes\n", *g->loaded, min_written);
	if (min_written) {
		*g->loaded -= min_written;
		FOR_EACH_WRITER_NODE(i, g)
			g->writer_nodes[i].written -= min_written;
	}
	if (!*g->loaded && *g->input_error) {
		g->error = *g->input_error;
		t->ret = g->error;
		return;
	}
	t->ret = 1;
	if (*g->loaded && min_written) {
//		PARA_INFO_LOG("moving %zd bytes\n", *g->loaded);
		memmove(g->buf, g->buf + min_written, *g->loaded);
	}
}

/**
 * call the open function of each writer in the group
 *
 * \param g the writer node group
 *
 * \return If at least one open function returned an error, all successful
 * writer notes get closed and this error value is returned. Upon success, a
 * task associated with \a g is registered to the scheduler and the function
 * returns a positive value.
 * */
int wng_open(struct writer_node_group *g)
{
	int i, ret = 1;

	PARA_NOTICE_LOG("opening wng %p with %d writer(s)\n", g, g->num_writers);
	register_task(&g->task);
	FOR_EACH_WRITER_NODE(i, g) {
		struct writer_node *wn = &g->writer_nodes[i];
		wn->wng = g;
		ret = wn->writer->open(wn);
		if (ret < 0)
			goto err_out;
		wn->chunk_bytes = ret;
		g->max_chunk_bytes = PARA_MAX(g->max_chunk_bytes, ret);
	}
	sprintf(g->task.status, "%s", "writer node group");
	g->error = 0;
	return 1;
err_out:
	PARA_ERROR_LOG("%s\n", para_strerror(-ret));
	unregister_task(&g->task);
	while (i > 0) {
		struct writer_node *wn = &g->writer_nodes[--i];
		wn->writer->close(wn);
	}
	g->num_writers = 0;
	g->error = ret;
	return ret;
}


/**
 * unregister a writer node group task
 *
 * \param g the group whose task is to be closed
 */
void wng_unregister(struct writer_node_group *g)
{
	unregister_task(&g->task);
}

/**
 * call the close function of each writer in the given group
 *
 * \param g the writer node group to close
 *
 * This function also frees all resources of the given group.
 */
void wng_close(struct writer_node_group *g)
{
	int i;

	if (!g)
		return;
	PARA_NOTICE_LOG("closing wng with %d writer(s)\n", g->num_writers);
	FOR_EACH_WRITER_NODE(i, g) {
		struct writer_node *wn = &g->writer_nodes[i];
		wn->writer->close(wn);
	}
	free(g->writer_nodes);
	free(g);
}

/**
 * allocate and initialize a new writer_node_group struct
 *
 * \param num_writers the number of writer nodes for the new group
 *
 * \return Pointer to the new writer node group
 */
struct writer_node_group *wng_new(unsigned num_writers)
{
	struct writer_node_group *g = para_calloc(sizeof(struct writer_node_group));
	g->num_writers = num_writers;
	g->writer_nodes = para_calloc(num_writers
		* sizeof(struct writer_node));
	g->task.private_data = g;
	g->task.post_select = wng_post_select;
	g->task.pre_select = wng_pre_select;
	return g;
}

/**
 * call the init function of each supported paraslash writer
 */
void init_supported_writers(void)
{
	int i;

	FOR_EACH_WRITER(i)
		writers[i].init(&writers[i]);
}
/**
 * check if given string is a valid command line for any writer
 *
 * \param \wa string of the form writer_name:options
 * \param writer_num contains the number of the writer upon success
 *
 * This function checks whether \a wa starts with the name of a supported
 * paraslash writer, optionally followed by a colon and any options for that
 * writer.  If a valid writer name was found and further are present, the
 * remaining part of \a wa is passed to that writer's config parser.
 *
 * \return On success, a pointer to the gengetopt args info struct is returned
 * and \a writer_num contains the number of the writer. Otherwise this function
 * returns \p NULL.
 */
void *check_writer_arg(const char *wa, int *writer_num)
{
	int i;

	*writer_num = -E_WRITE_COMMON_SYNTAX;
	PARA_INFO_LOG("checking  %s\n", wa);
	FOR_EACH_WRITER(i) {
		const char *name = writer_names[i];
		size_t len = strlen(name);
		char c;
		if (strlen(wa) < len)
			continue;
		if (strncmp(name, wa, len))
			continue;
		c = wa[len];
		if (c && c != ' ')
			continue;
		if (c && !writers[i].parse_config)
			return NULL;
		*writer_num = i;
		return writers[i].parse_config(c? wa + len + 1 : "");
	}
	PARA_ERROR_LOG("writer not found\n");
	return NULL;
}

/**
 * setup a writer node group with only one writer, the default writer
 *
 * The writer which is set up depends on the OS. It defaults to alsa for Linux,
 * osx_write for OS X, file writer if neither of these is supported.
 *
 * \return pointer to the allocated writer node group
 */
struct writer_node_group *setup_default_wng(void)
{
	struct writer_node_group *wng = wng_new(1);
	wng->writer_nodes[0].writer = &writers[DEFAULT_WRITER];
	PARA_INFO_LOG("using default writer: %s %p\n",
		writer_names[DEFAULT_WRITER], writers[DEFAULT_WRITER].parse_config);
	wng->writer_nodes[0].conf = writers[DEFAULT_WRITER].parse_config("");
	return wng;
}
